@@ -31,6 +31,7 @@
 #include "src/inspector/injected-script.h"
 
 #include <cmath>
+#include <memory>
 #include <unordered_set>
 
 #include "../../third_party/inspector_protocol/crdtp/json.h"
@@ -46,8 +47,10 @@
 #include "src/inspector/remote-object-id.h"
 #include "src/inspector/string-util.h"
 #include "src/inspector/v8-console.h"
+#include "src/inspector/v8-debugger.h"
 #include "src/inspector/v8-inspector-impl.h"
 #include "src/inspector/v8-inspector-session-impl.h"
+#include "src/inspector/v8-serialization-duplicate-tracker.h"
 #include "src/inspector/v8-stack-trace-impl.h"
 #include "src/inspector/v8-value-utils.h"
 #include "src/inspector/value-mirror.h"
@@ -96,29 +99,13 @@ class InjectedScript::ProtocolPromiseHandler {
   static void add(V8InspectorSessionImpl* session,
                   v8::Local<v8::Context> context, v8::Local<v8::Value> value,
                   int executionContextId, const String16& objectGroup,
-                  WrapMode wrapMode, bool replMode, bool throwOnSideEffect,
+                  std::unique_ptr<WrapOptions> wrapOptions, bool replMode,
+                  bool throwOnSideEffect,
                   std::weak_ptr<EvaluateCallback> callback) {
     InjectedScript::ContextScope scope(session, executionContextId);
     Response response = scope.initialize();
     if (!response.IsSuccess()) return;
 
-    v8::MaybeLocal<v8::Promise> originalPromise =
-        value->IsPromise() ? value.As<v8::Promise>()
-                           : v8::MaybeLocal<v8::Promise>();
-    V8InspectorImpl* inspector = session->inspector();
-    ProtocolPromiseHandler* handler = new ProtocolPromiseHandler(
-        session, executionContextId, objectGroup, wrapMode, replMode,
-        throwOnSideEffect, callback, originalPromise);
-    v8::Local<v8::Value> wrapper = handler->m_wrapper.Get(inspector->isolate());
-    v8::Local<v8::Function> thenCallbackFunction =
-        v8::Function::New(context, thenCallback, wrapper, 0,
-                          v8::ConstructorBehavior::kThrow)
-            .ToLocalChecked();
-    v8::Local<v8::Function> catchCallbackFunction =
-        v8::Function::New(context, catchCallback, wrapper, 0,
-                          v8::ConstructorBehavior::kThrow)
-            .ToLocalChecked();
-
     v8::Local<v8::Promise> promise;
     v8::Local<v8::Promise::Resolver> resolver;
     if (value->IsPromise()) {
@@ -140,6 +127,22 @@ class InjectedScript::ProtocolPromiseHandler {
       promise = resolver->GetPromise();
     }
 
+    V8InspectorImpl* inspector = session->inspector();
+    PromiseHandlerTracker::Id handlerId =
+        inspector->promiseHandlerTracker().create(
+            session, executionContextId, objectGroup, std::move(wrapOptions),
+            replMode, throwOnSideEffect, callback, promise);
+    v8::Local<v8::Number> data =
+        v8::Number::New(inspector->isolate(), handlerId);
+    v8::Local<v8::Function> thenCallbackFunction =
+        v8::Function::New(context, thenCallback, data, 0,
+                          v8::ConstructorBehavior::kThrow)
+            .ToLocalChecked();
+    v8::Local<v8::Function> catchCallbackFunction =
+        v8::Function::New(context, catchCallback, data, 0,
+                          v8::ConstructorBehavior::kThrow)
+            .ToLocalChecked();
+
     if (promise->Then(context, thenCallbackFunction, catchCallbackFunction)
             .IsEmpty()) {
       // Re-initialize after returning from JS.
@@ -151,6 +154,8 @@ class InjectedScript::ProtocolPromiseHandler {
   }
 
  private:
+  friend class PromiseHandlerTracker;
+
   static v8::Local<v8::String> GetDotReplResultString(v8::Isolate* isolate) {
     // TODO(szuend): Cache the string in a v8::Persistent handle.
     return v8::String::NewFromOneByte(
@@ -159,65 +164,84 @@ class InjectedScript::ProtocolPromiseHandler {
   }
 
   static void thenCallback(const v8::FunctionCallbackInfo<v8::Value>& info) {
-    ProtocolPromiseHandler* handler = static_cast<ProtocolPromiseHandler*>(
-        info.Data().As<v8::External>()->Value());
-    DCHECK(handler);
+    PromiseHandlerTracker::Id handlerId =
+        static_cast<PromiseHandlerTracker::Id>(
+            info.Data().As<v8::Number>()->Value());
+    PromiseHandlerTracker& handlerTracker =
+        static_cast<V8InspectorImpl*>(
+            v8::debug::GetInspector(info.GetIsolate()))
+            ->promiseHandlerTracker();
+    // We currently store the handlers with the inspector. In rare cases the
+    // inspector dies (discarding the handler) with the micro task queue
+    // running after. Don't do anything in that case.
+    ProtocolPromiseHandler* handler = handlerTracker.get(handlerId);
+    if (!handler) return;
     v8::Local<v8::Value> value =
         info.Length() > 0 ? info[0]
                           : v8::Undefined(info.GetIsolate()).As<v8::Value>();
     handler->thenCallback(value);
-    delete handler;
+    handlerTracker.discard(handlerId,
+                           PromiseHandlerTracker::DiscardReason::kFulfilled);
   }
 
   static void catchCallback(const v8::FunctionCallbackInfo<v8::Value>& info) {
-    ProtocolPromiseHandler* handler = static_cast<ProtocolPromiseHandler*>(
-        info.Data().As<v8::External>()->Value());
-    DCHECK(handler);
+    PromiseHandlerTracker::Id handlerId =
+        static_cast<PromiseHandlerTracker::Id>(
+            info.Data().As<v8::Number>()->Value());
+    PromiseHandlerTracker& handlerTracker =
+        static_cast<V8InspectorImpl*>(
+            v8::debug::GetInspector(info.GetIsolate()))
+            ->promiseHandlerTracker();
+    // We currently store the handlers with the inspector. In rare cases the
+    // inspector dies (discarding the handler) with the micro task queue
+    // running after. Don't do anything in that case.
+    ProtocolPromiseHandler* handler = handlerTracker.get(handlerId);
+    if (!handler) return;
     v8::Local<v8::Value> value =
         info.Length() > 0 ? info[0]
                           : v8::Undefined(info.GetIsolate()).As<v8::Value>();
     handler->catchCallback(value);
-    delete handler;
+    handlerTracker.discard(handlerId,
+                           PromiseHandlerTracker::DiscardReason::kFulfilled);
   }
 
-  ProtocolPromiseHandler(V8InspectorSessionImpl* session,
+  ProtocolPromiseHandler(PromiseHandlerTracker::Id id,
+                         V8InspectorSessionImpl* session,
                          int executionContextId, const String16& objectGroup,
-                         WrapMode wrapMode, bool replMode,
-                         bool throwOnSideEffect,
+                         std::unique_ptr<WrapOptions> wrapOptions,
+                         bool replMode, bool throwOnSideEffect,
                          std::weak_ptr<EvaluateCallback> callback,
-                         v8::MaybeLocal<v8::Promise> maybeEvaluationResult)
+                         v8::Local<v8::Promise> evaluationResult)
       : m_inspector(session->inspector()),
         m_sessionId(session->sessionId()),
         m_contextGroupId(session->contextGroupId()),
         m_executionContextId(executionContextId),
         m_objectGroup(objectGroup),
-        m_wrapMode(wrapMode),
+        m_wrapOptions(std::move(wrapOptions)),
         m_replMode(replMode),
         m_throwOnSideEffect(throwOnSideEffect),
         m_callback(std::move(callback)),
-        m_wrapper(m_inspector->isolate(),
-                  v8::External::New(m_inspector->isolate(), this)) {
-    m_wrapper.SetWeak(this, cleanup, v8::WeakCallbackType::kParameter);
-    v8::Local<v8::Promise> promise;
-    if (maybeEvaluationResult.ToLocal(&promise)) {
-      m_evaluationResult =
-          v8::Global<v8::Promise>(m_inspector->isolate(), promise);
-    }
+        m_evaluationResult(m_inspector->isolate(), evaluationResult) {
+    m_evaluationResult.SetWeak(reinterpret_cast<PromiseHandlerTracker::Id*>(id),
+                               cleanup, v8::WeakCallbackType::kParameter);
   }
 
   static void cleanup(
-      const v8::WeakCallbackInfo<ProtocolPromiseHandler>& data) {
-    if (!data.GetParameter()->m_wrapper.IsEmpty()) {
-      data.GetParameter()->m_wrapper.Reset();
-      data.GetParameter()->m_evaluationResult.Reset();
-      data.SetSecondPassCallback(cleanup);
-    } else {
-      data.GetParameter()->sendPromiseCollected();
-      delete data.GetParameter();
-    }
+      const v8::WeakCallbackInfo<PromiseHandlerTracker::Id>& data) {
+    auto id = reinterpret_cast<PromiseHandlerTracker::Id>(data.GetParameter());
+    PromiseHandlerTracker& handlerTracker =
+        static_cast<V8InspectorImpl*>(
+            v8::debug::GetInspector(data.GetIsolate()))
+            ->promiseHandlerTracker();
+    // {discard} deletes the {ProtocolPromiseHandler} which resets the handle.
+    handlerTracker.discard(
+        id, PromiseHandlerTracker::DiscardReason::kPromiseCollected);
   }
 
   void thenCallback(v8::Local<v8::Value> value) {
+    // We don't need the m_evaluationResult in the `thenCallback`, but we also
+    // don't want `cleanup` running in case we re-enter JS.
+    m_evaluationResult.Reset();
     V8InspectorSessionImpl* session =
         m_inspector->sessionById(m_contextGroupId, m_sessionId);
     if (!session) return;
@@ -250,8 +274,8 @@ class InjectedScript::ProtocolPromiseHandler {
     }
 
     std::unique_ptr<protocol::Runtime::RemoteObject> wrappedValue;
-    response = scope.injectedScript()->wrapObject(result, m_objectGroup,
-                                                  m_wrapMode, &wrappedValue);
+    response = scope.injectedScript()->wrapObject(
+        result, m_objectGroup, *m_wrapOptions, &wrappedValue);
     if (!response.IsSuccess()) {
       EvaluateCallback::sendFailure(m_callback, scope.injectedScript(),
                                     response);
@@ -263,6 +287,9 @@ class InjectedScript::ProtocolPromiseHandler {
   }
 
   void catchCallback(v8::Local<v8::Value> result) {
+    // Hold strongly onto m_evaluationResult now to prevent `cleanup` from
+    // running in case any code below triggers GC.
+    m_evaluationResult.ClearWeak();
     V8InspectorSessionImpl* session =
         m_inspector->sessionById(m_contextGroupId, m_sessionId);
     if (!session) return;
@@ -270,8 +297,8 @@ class InjectedScript::ProtocolPromiseHandler {
     Response response = scope.initialize();
     if (!response.IsSuccess()) return;
     std::unique_ptr<protocol::Runtime::RemoteObject> wrappedValue;
-    response = scope.injectedScript()->wrapObject(result, m_objectGroup,
-                                                  m_wrapMode, &wrappedValue);
+    response = scope.injectedScript()->wrapObject(
+        result, m_objectGroup, *m_wrapOptions, &wrappedValue);
     if (!response.IsSuccess()) {
       EvaluateCallback::sendFailure(m_callback, scope.injectedScript(),
                                     response);
@@ -360,28 +387,15 @@ class InjectedScript::ProtocolPromiseHandler {
                                   std::move(exceptionDetails));
   }
 
-  void sendPromiseCollected() {
-    V8InspectorSessionImpl* session =
-        m_inspector->sessionById(m_contextGroupId, m_sessionId);
-    if (!session) return;
-    InjectedScript::ContextScope scope(session, m_executionContextId);
-    Response response = scope.initialize();
-    if (!response.IsSuccess()) return;
-    EvaluateCallback::sendFailure(
-        m_callback, scope.injectedScript(),
-        Response::ServerError("Promise was collected"));
-  }
-
   V8InspectorImpl* m_inspector;
   int m_sessionId;
   int m_contextGroupId;
   int m_executionContextId;
   String16 m_objectGroup;
-  WrapMode m_wrapMode;
+  std::unique_ptr<WrapOptions> m_wrapOptions;
   bool m_replMode;
   bool m_throwOnSideEffect;
   std::weak_ptr<EvaluateCallback> m_callback;
-  v8::Global<v8::External> m_wrapper;
   v8::Global<v8::Promise> m_evaluationResult;
 };
 
@@ -408,7 +422,8 @@ class PropertyAccumulator : public ValueMirror::PropertyAccumulator {
 Response InjectedScript::getProperties(
     v8::Local<v8::Object> object, const String16& groupName, bool ownProperties,
     bool accessorPropertiesOnly, bool nonIndexedPropertiesOnly,
-    WrapMode wrapMode, std::unique_ptr<Array<PropertyDescriptor>>* properties,
+    const WrapOptions& wrapOptions,
+    std::unique_ptr<Array<PropertyDescriptor>>* properties,
     Maybe<protocol::Runtime::ExceptionDetails>* exceptionDetails) {
   v8::HandleScope handles(m_context->isolate());
   v8::Local<v8::Context> context = m_context->context();
@@ -435,7 +450,7 @@ Response InjectedScript::getProperties(
     std::unique_ptr<RemoteObject> remoteObject;
     if (mirror.value) {
       Response response = wrapObjectMirror(
-          *mirror.value, groupName, wrapMode, v8::MaybeLocal<v8::Value>(),
+          *mirror.value, groupName, wrapOptions, v8::MaybeLocal<v8::Value>(),
           kMaxCustomPreviewDepth, &remoteObject);
       if (!response.IsSuccess()) return response;
       descriptor->setValue(std::move(remoteObject));
@@ -443,40 +458,40 @@ Response InjectedScript::getProperties(
     }
     if (mirror.getter) {
       Response response =
-          mirror.getter->buildRemoteObject(context, wrapMode, &remoteObject);
+          mirror.getter->buildRemoteObject(context, wrapOptions, &remoteObject);
       if (!response.IsSuccess()) return response;
-      response =
-          bindRemoteObjectIfNeeded(sessionId, context, mirror.getter->v8Value(),
-                                   groupName, remoteObject.get());
+      response = bindRemoteObjectIfNeeded(sessionId, context,
+                                          mirror.getter->v8Value(isolate),
+                                          groupName, remoteObject.get());
       if (!response.IsSuccess()) return response;
       descriptor->setGet(std::move(remoteObject));
     }
     if (mirror.setter) {
       Response response =
-          mirror.setter->buildRemoteObject(context, wrapMode, &remoteObject);
+          mirror.setter->buildRemoteObject(context, wrapOptions, &remoteObject);
       if (!response.IsSuccess()) return response;
-      response =
-          bindRemoteObjectIfNeeded(sessionId, context, mirror.setter->v8Value(),
-                                   groupName, remoteObject.get());
+      response = bindRemoteObjectIfNeeded(sessionId, context,
+                                          mirror.setter->v8Value(isolate),
+                                          groupName, remoteObject.get());
       if (!response.IsSuccess()) return response;
       descriptor->setSet(std::move(remoteObject));
     }
     if (mirror.symbol) {
       Response response =
-          mirror.symbol->buildRemoteObject(context, wrapMode, &remoteObject);
+          mirror.symbol->buildRemoteObject(context, wrapOptions, &remoteObject);
       if (!response.IsSuccess()) return response;
-      response =
-          bindRemoteObjectIfNeeded(sessionId, context, mirror.symbol->v8Value(),
-                                   groupName, remoteObject.get());
+      response = bindRemoteObjectIfNeeded(sessionId, context,
+                                          mirror.symbol->v8Value(isolate),
+                                          groupName, remoteObject.get());
       if (!response.IsSuccess()) return response;
       descriptor->setSymbol(std::move(remoteObject));
     }
     if (mirror.exception) {
-      Response response =
-          mirror.exception->buildRemoteObject(context, wrapMode, &remoteObject);
+      Response response = mirror.exception->buildRemoteObject(
+          context, wrapOptions, &remoteObject);
       if (!response.IsSuccess()) return response;
       response = bindRemoteObjectIfNeeded(sessionId, context,
-                                          mirror.exception->v8Value(),
+                                          mirror.exception->v8Value(isolate),
                                           groupName, remoteObject.get());
       if (!response.IsSuccess()) return response;
       descriptor->setValue(std::move(remoteObject));
@@ -511,11 +526,13 @@ Response InjectedScript::getInternalAndPrivateProperties(
     for (const auto& internalProperty : internalPropertiesWrappers) {
       std::unique_ptr<RemoteObject> remoteObject;
       Response response = internalProperty.value->buildRemoteObject(
-          m_context->context(), WrapMode::kNoPreview, &remoteObject);
+          m_context->context(), WrapOptions({WrapMode::kIdOnly}),
+          &remoteObject);
       if (!response.IsSuccess()) return response;
-      response = bindRemoteObjectIfNeeded(sessionId, context,
-                                          internalProperty.value->v8Value(),
-                                          groupName, remoteObject.get());
+      response = bindRemoteObjectIfNeeded(
+          sessionId, context,
+          internalProperty.value->v8Value(context->GetIsolate()), groupName,
+          remoteObject.get());
       if (!response.IsSuccess()) return response;
       (*internalProperties)
           ->emplace_back(InternalPropertyDescriptor::create()
@@ -539,33 +556,36 @@ Response InjectedScript::getInternalAndPrivateProperties(
            (!!privateProperty.value));
     if (privateProperty.value) {
       Response response = privateProperty.value->buildRemoteObject(
-          context, WrapMode::kNoPreview, &remoteObject);
+          context, WrapOptions({WrapMode::kIdOnly}), &remoteObject);
       if (!response.IsSuccess()) return response;
-      response = bindRemoteObjectIfNeeded(sessionId, context,
-                                          privateProperty.value->v8Value(),
-                                          groupName, remoteObject.get());
+      response = bindRemoteObjectIfNeeded(
+          sessionId, context,
+          privateProperty.value->v8Value(context->GetIsolate()), groupName,
+          remoteObject.get());
       if (!response.IsSuccess()) return response;
       descriptor->setValue(std::move(remoteObject));
     }
 
     if (privateProperty.getter) {
       Response response = privateProperty.getter->buildRemoteObject(
-          context, WrapMode::kNoPreview, &remoteObject);
+          context, WrapOptions({WrapMode::kIdOnly}), &remoteObject);
       if (!response.IsSuccess()) return response;
-      response = bindRemoteObjectIfNeeded(sessionId, context,
-                                          privateProperty.getter->v8Value(),
-                                          groupName, remoteObject.get());
+      response = bindRemoteObjectIfNeeded(
+          sessionId, context,
+          privateProperty.getter->v8Value(context->GetIsolate()), groupName,
+          remoteObject.get());
       if (!response.IsSuccess()) return response;
       descriptor->setGet(std::move(remoteObject));
     }
 
     if (privateProperty.setter) {
       Response response = privateProperty.setter->buildRemoteObject(
-          context, WrapMode::kNoPreview, &remoteObject);
+          context, WrapOptions({WrapMode::kIdOnly}), &remoteObject);
       if (!response.IsSuccess()) return response;
-      response = bindRemoteObjectIfNeeded(sessionId, context,
-                                          privateProperty.setter->v8Value(),
-                                          groupName, remoteObject.get());
+      response = bindRemoteObjectIfNeeded(
+          sessionId, context,
+          privateProperty.setter->v8Value(context->GetIsolate()), groupName,
+          remoteObject.get());
       if (!response.IsSuccess()) return response;
       descriptor->setSet(std::move(remoteObject));
     }
@@ -582,35 +602,38 @@ void InjectedScript::releaseObject(const String16& objectId) {
 }
 
 Response InjectedScript::wrapObject(
-    v8::Local<v8::Value> value, const String16& groupName, WrapMode wrapMode,
+    v8::Local<v8::Value> value, const String16& groupName,
+    const WrapOptions& wrapOptions,
     std::unique_ptr<protocol::Runtime::RemoteObject>* result) {
-  return wrapObject(value, groupName, wrapMode, v8::MaybeLocal<v8::Value>(),
+  return wrapObject(value, groupName, wrapOptions, v8::MaybeLocal<v8::Value>(),
                     kMaxCustomPreviewDepth, result);
 }
 
 Response InjectedScript::wrapObject(
-    v8::Local<v8::Value> value, const String16& groupName, WrapMode wrapMode,
+    v8::Local<v8::Value> value, const String16& groupName,
+    const WrapOptions& wrapOptions,
     v8::MaybeLocal<v8::Value> customPreviewConfig, int maxCustomPreviewDepth,
     std::unique_ptr<protocol::Runtime::RemoteObject>* result) {
   v8::Local<v8::Context> context = m_context->context();
   v8::Context::Scope contextScope(context);
   std::unique_ptr<ValueMirror> mirror = ValueMirror::create(context, value);
   if (!mirror) return Response::InternalError();
-  return wrapObjectMirror(*mirror, groupName, wrapMode, customPreviewConfig,
+  return wrapObjectMirror(*mirror, groupName, wrapOptions, customPreviewConfig,
                           maxCustomPreviewDepth, result);
 }
 
 Response InjectedScript::wrapObjectMirror(
-    const ValueMirror& mirror, const String16& groupName, WrapMode wrapMode,
+    const ValueMirror& mirror, const String16& groupName,
+    const WrapOptions& wrapOptions,
     v8::MaybeLocal<v8::Value> customPreviewConfig, int maxCustomPreviewDepth,
     std::unique_ptr<protocol::Runtime::RemoteObject>* result) {
   int customPreviewEnabled = m_customPreviewEnabled;
   int sessionId = m_sessionId;
   v8::Local<v8::Context> context = m_context->context();
   v8::Context::Scope contextScope(context);
-  Response response = mirror.buildRemoteObject(context, wrapMode, result);
+  Response response = mirror.buildRemoteObject(context, wrapOptions, result);
   if (!response.IsSuccess()) return response;
-  v8::Local<v8::Value> value = mirror.v8Value();
+  v8::Local<v8::Value> value = mirror.v8Value(context->GetIsolate());
   response = bindRemoteObjectIfNeeded(sessionId, context, value, groupName,
                                       result->get());
   if (!response.IsSuccess()) return response;
@@ -621,12 +644,39 @@ Response InjectedScript::wrapObjectMirror(
                           &customPreview);
     if (customPreview) (*result)->setCustomPreview(std::move(customPreview));
   }
-  if (wrapMode == WrapMode::kGenerateWebDriverValue) {
-    int maxDepth = 1;
-    std::unique_ptr<protocol::Runtime::WebDriverValue> webDriverValue =
-        mirror.buildWebDriverValue(context, maxDepth);
+  if (wrapOptions.mode == WrapMode::kDeep) {
+    V8SerializationDuplicateTracker duplicateTracker{context};
+
+    std::unique_ptr<protocol::DictionaryValue> deepSerializedValueDict;
+    response = mirror.buildDeepSerializedValue(
+        context, wrapOptions.serializationOptions.maxDepth,
+        wrapOptions.serializationOptions.additionalParameters.Get(
+            m_context->isolate()),
+        duplicateTracker, &deepSerializedValueDict);
     if (!response.IsSuccess()) return response;
-    (*result)->setWebDriverValue(std::move(webDriverValue));
+
+    String16 type;
+    deepSerializedValueDict->getString("type", &type);
+
+    std::unique_ptr<protocol::Runtime::DeepSerializedValue>
+        deepSerializedValue = protocol::Runtime::DeepSerializedValue::create()
+                                  .setType(type)
+                                  .build();
+
+    protocol::Value* maybeValue = deepSerializedValueDict->get("value");
+    if (maybeValue != nullptr) {
+      deepSerializedValue->setValue(maybeValue->clone());
+    }
+
+    int weakLocalObjectReference;
+    if (deepSerializedValueDict->getInteger("weakLocalObjectReference",
+                                            &weakLocalObjectReference)) {
+      deepSerializedValue->setWeakLocalObjectReference(
+          weakLocalObjectReference);
+    }
+
+    if (!response.IsSuccess()) return response;
+    (*result)->setDeepSerializedValue(std::move(deepSerializedValue));
   }
 
   return Response::Success();
@@ -634,18 +684,18 @@ Response InjectedScript::wrapObjectMirror(
 
 std::unique_ptr<protocol::Runtime::RemoteObject> InjectedScript::wrapTable(
     v8::Local<v8::Object> table, v8::MaybeLocal<v8::Array> maybeColumns) {
-  using protocol::Runtime::RemoteObject;
+  using protocol::Array;
   using protocol::Runtime::ObjectPreview;
   using protocol::Runtime::PropertyPreview;
-  using protocol::Array;
+  using protocol::Runtime::RemoteObject;
 
   v8::Isolate* isolate = m_context->isolate();
   v8::HandleScope handles(isolate);
   v8::Local<v8::Context> context = m_context->context();
 
   std::unique_ptr<RemoteObject> remoteObject;
-  Response response =
-      wrapObject(table, "console", WrapMode::kNoPreview, &remoteObject);
+  Response response = wrapObject(
+      table, "console", WrapOptions({WrapMode::kIdOnly}), &remoteObject);
   if (!remoteObject || !response.IsSuccess()) return nullptr;
 
   auto mirror = ValueMirror::create(context, table);
@@ -697,8 +747,9 @@ std::unique_ptr<protocol::Runtime::RemoteObject> InjectedScript::wrapTable(
 
 void InjectedScript::addPromiseCallback(
     V8InspectorSessionImpl* session, v8::MaybeLocal<v8::Value> value,
-    const String16& objectGroup, WrapMode wrapMode, bool replMode,
-    bool throwOnSideEffect, std::shared_ptr<EvaluateCallback> callback) {
+    const String16& objectGroup, std::unique_ptr<WrapOptions> wrapOptions,
+    bool replMode, bool throwOnSideEffect,
+    std::shared_ptr<EvaluateCallback> callback) {
   m_evaluateCallbacks.insert(callback);
   // After stashing the shared_ptr in `m_evaluateCallback`, we reset `callback`.
   // `ProtocolPromiseHandler:add` can take longer than the life time of this
@@ -717,7 +768,7 @@ void InjectedScript::addPromiseCallback(
                                       v8::MicrotasksScope::kRunMicrotasks);
   ProtocolPromiseHandler::add(session, m_context->context(),
                               value.ToLocalChecked(), m_context->contextId(),
-                              objectGroup, wrapMode, replMode,
+                              objectGroup, std::move(wrapOptions), replMode,
                               throwOnSideEffect, weak_callback);
   // Do not add any code here! `this` might be invalid.
   // `ProtocolPromiseHandler::add` calls into JS which could kill this
@@ -835,8 +886,8 @@ Response InjectedScript::addExceptionToDetails(
   std::unique_ptr<protocol::Runtime::RemoteObject> wrapped;
   Response response =
       wrapObject(exception, objectGroup,
-                 exception->IsNativeError() ? WrapMode::kNoPreview
-                                            : WrapMode::kWithPreview,
+                 exception->IsNativeError() ? WrapOptions({WrapMode::kIdOnly})
+                                            : WrapOptions({WrapMode::kPreview}),
                  &wrapped);
   if (!response.IsSuccess()) return response;
   exceptionDetails->setException(std::move(wrapped));
@@ -896,14 +947,20 @@ Response InjectedScript::createExceptionDetails(
 
 Response InjectedScript::wrapEvaluateResult(
     v8::MaybeLocal<v8::Value> maybeResultValue, const v8::TryCatch& tryCatch,
-    const String16& objectGroup, WrapMode wrapMode, bool throwOnSideEffect,
+    const String16& objectGroup, const WrapOptions& wrapOptions,
+    bool throwOnSideEffect,
     std::unique_ptr<protocol::Runtime::RemoteObject>* result,
     Maybe<protocol::Runtime::ExceptionDetails>* exceptionDetails) {
   v8::Local<v8::Value> resultValue;
   if (!tryCatch.HasCaught()) {
-    if (!maybeResultValue.ToLocal(&resultValue))
+    if (!maybeResultValue.ToLocal(&resultValue)) {
+      if (!tryCatch.CanContinue()) {
+        return Response::ServerError("Execution was terminated");
+      }
       return Response::InternalError();
-    Response response = wrapObject(resultValue, objectGroup, wrapMode, result);
+    }
+    Response response =
+        wrapObject(resultValue, objectGroup, wrapOptions, result);
     if (!response.IsSuccess()) return response;
     if (objectGroup == "console") {
       m_lastEvaluationResult.Reset(m_context->isolate(), resultValue);
@@ -918,11 +975,11 @@ Response InjectedScript::wrapEvaluateResult(
       m_context->inspector()->client()->dispatchError(
           m_context->context(), tryCatch.Message(), exception);
     }
-    Response response =
-        wrapObject(exception, objectGroup,
-                   exception->IsNativeError() ? WrapMode::kNoPreview
-                                              : WrapMode::kWithPreview,
-                   result);
+    Response response = wrapObject(exception, objectGroup,
+                                   exception->IsNativeError()
+                                       ? WrapOptions({WrapMode::kIdOnly})
+                                       : WrapOptions({WrapMode::kPreview}),
+                                   result);
     if (!response.IsSuccess()) return response;
     // We send exception in result for compatibility reasons, even though it's
     // accessible through exceptionDetails.exception.
@@ -1013,6 +1070,10 @@ void InjectedScript::Scope::allowCodeGenerationFromStrings() {
   m_context->AllowCodeGenerationFromStrings(true);
 }
 
+void InjectedScript::Scope::setTryCatchVerbose() {
+  m_tryCatch.SetVerbose(true);
+}
+
 void InjectedScript::Scope::cleanup() {
   m_commandLineAPIScope.reset();
   if (!m_context.IsEmpty()) {
@@ -1123,4 +1184,68 @@ void InjectedScript::unbindObject(int id) {
   m_idToObjectGroupName.erase(id);
 }
 
+PromiseHandlerTracker::PromiseHandlerTracker() = default;
+
+PromiseHandlerTracker::~PromiseHandlerTracker() { discardAll(); }
+
+template <typename... Args>
+PromiseHandlerTracker::Id PromiseHandlerTracker::create(Args&&... args) {
+  Id id = m_lastUsedId++;
+  InjectedScript::ProtocolPromiseHandler* handler =
+      new InjectedScript::ProtocolPromiseHandler(id,
+                                                 std::forward<Args>(args)...);
+  m_promiseHandlers.emplace(id, handler);
+  return id;
+}
+
+void PromiseHandlerTracker::discard(Id id, DiscardReason reason) {
+  auto iter = m_promiseHandlers.find(id);
+  CHECK_NE(iter, m_promiseHandlers.end());
+  InjectedScript::ProtocolPromiseHandler* handler = iter->second.get();
+
+  switch (reason) {
+    case DiscardReason::kPromiseCollected:
+      sendFailure(handler, Response::ServerError("Promise was collected"));
+      break;
+    case DiscardReason::kTearDown:
+      sendFailure(handler, Response::ServerError(
+                               "Tearing down inspector/session/context"));
+      break;
+    case DiscardReason::kFulfilled:
+      // Do nothing.
+      break;
+  }
+
+  m_promiseHandlers.erase(id);
+}
+
+InjectedScript::ProtocolPromiseHandler* PromiseHandlerTracker::get(
+    Id id) const {
+  auto iter = m_promiseHandlers.find(id);
+  if (iter == m_promiseHandlers.end()) return nullptr;
+
+  return iter->second.get();
+}
+
+void PromiseHandlerTracker::sendFailure(
+    InjectedScript::ProtocolPromiseHandler* handler,
+    const protocol::DispatchResponse& response) const {
+  V8InspectorImpl* inspector = handler->m_inspector;
+  V8InspectorSessionImpl* session =
+      inspector->sessionById(handler->m_contextGroupId, handler->m_sessionId);
+  if (!session) return;
+  InjectedScript::ContextScope scope(session, handler->m_executionContextId);
+  Response res = scope.initialize();
+  if (!res.IsSuccess()) return;
+  EvaluateCallback::sendFailure(handler->m_callback, scope.injectedScript(),
+                                response);
+}
+
+void PromiseHandlerTracker::discardAll() {
+  while (!m_promiseHandlers.empty()) {
+    discard(m_promiseHandlers.begin()->first, DiscardReason::kTearDown);
+  }
+  CHECK(m_promiseHandlers.empty());
+}
+
 }  // namespace v8_inspector
