@@ -83,9 +83,10 @@ void LoopBuilder::JumpToHeader(int loop_depth, LoopBuilder* const parent_loop) {
     //
     // The loop must have closed form, i.e. all loop elements are within the
     // loop, the loop header precedes the body and next elements in the loop.
-    builder()->JumpLoop(&loop_header_,
-                        std::min(loop_depth, BytecodeArray::kMaxOsrUrgency - 1),
-                        source_position_);
+    int slot_index = feedback_vector_spec_->AddJumpLoopSlot().ToInt();
+    builder()->JumpLoop(
+        &loop_header_, std::min(loop_depth, FeedbackVector::kMaxOsrUrgency - 1),
+        source_position_, slot_index);
   }
 }
 
@@ -207,6 +208,43 @@ void TryFinallyBuilder::EndFinally() {
   // Nothing to be done here.
 }
 
+ConditionalChainControlFlowBuilder::~ConditionalChainControlFlowBuilder() {
+  end_labels_.Bind(builder());
+#ifdef DEBUG
+  DCHECK(end_labels_.empty() || end_labels_.is_bound());
+
+  for (auto* label : then_labels_list_) {
+    DCHECK(label->empty() || label->is_bound());
+  }
+
+  for (auto* label : else_labels_list_) {
+    DCHECK(label->empty() || label->is_bound());
+  }
+#endif
+}
+
+void ConditionalChainControlFlowBuilder::JumpToEnd() {
+  builder()->Jump(end_labels_.New());
+}
+
+void ConditionalChainControlFlowBuilder::ThenAt(size_t index) {
+  DCHECK_LT(index, then_labels_list_.length());
+  then_labels_at(index)->Bind(builder());
+  if (block_coverage_builder_) {
+    block_coverage_builder_->IncrementBlockCounter(
+        block_coverage_then_slot_at(index));
+  }
+}
+
+void ConditionalChainControlFlowBuilder::ElseAt(size_t index) {
+  DCHECK_LT(index, else_labels_list_.length());
+  else_labels_at(index)->Bind(builder());
+  if (block_coverage_builder_) {
+    block_coverage_builder_->IncrementBlockCounter(
+        block_coverage_else_slot_at(index));
+  }
+}
+
 ConditionalControlFlowBuilder::~ConditionalControlFlowBuilder() {
   if (!else_labels_.is_bound()) else_labels_.Bind(builder());
   end_labels_.Bind(builder());
