@@ -13,18 +13,17 @@
 namespace v8 {
 namespace internal {
 
-#include "torque-generated/src/objects/primitive-heap-object-tq.inc"
-
 // An abstract superclass for classes representing JavaScript primitive values
 // other than Smi. It doesn't carry any functionality but allows primitive
 // classes to be identified in the type system.
-class PrimitiveHeapObject
-    : public TorqueGeneratedPrimitiveHeapObject<PrimitiveHeapObject,
-                                                HeapObject> {
+V8_OBJECT class PrimitiveHeapObject : public HeapObjectLayout {
  public:
-  static_assert(kHeaderSize == HeapObject::kHeaderSize);
-  TQ_OBJECT_CONSTRUCTORS(PrimitiveHeapObject)
-};
+  DECL_CAST(PrimitiveHeapObject)
+  DECL_VERIFIER(PrimitiveHeapObject)
+} V8_OBJECT_END;
+
+static_assert(sizeof(PrimitiveHeapObject) == sizeof(HeapObjectLayout));
+static_assert(is_subtype_v<PrimitiveHeapObject, HeapObject>);
 
 }  // namespace internal
 }  // namespace v8
