@@ -28,15 +28,16 @@ const {
   ArrayPrototypeJoin,
   ArrayPrototypeLastIndexOf,
   ArrayPrototypePush,
+  ArrayPrototypePushApply,
   ArrayPrototypeSlice,
   ArrayPrototypeSort,
   ArrayPrototypeSplice,
   ArrayPrototypeUnshift,
-  ArrayPrototypePushApply,
   NumberIsInteger,
   ObjectAssign,
   ObjectDefineProperty,
   ObjectPrototypeHasOwnProperty,
+  PromiseWithResolvers,
   RegExpPrototypeExec,
   SafeSet,
   StringPrototypeIncludes,
@@ -45,8 +46,8 @@ const {
 } = primordials;
 
 const {
+  assignFunctionName,
   convertToValidSignal,
-  createDeferredPromise,
   getSystemErrorName,
   kEmptyObject,
   promisify,
@@ -64,16 +65,15 @@ const { Pipe, constants: PipeConstants } = internalBinding('pipe_wrap');
 
 const {
   AbortError,
-  codes: errorCodes,
+  codes: {
+    ERR_CHILD_PROCESS_IPC_REQUIRED,
+    ERR_CHILD_PROCESS_STDIO_MAXBUFFER,
+    ERR_INVALID_ARG_TYPE,
+    ERR_INVALID_ARG_VALUE,
+    ERR_OUT_OF_RANGE,
+  },
   genericNodeError,
 } = require('internal/errors');
-const {
-  ERR_INVALID_ARG_VALUE,
-  ERR_CHILD_PROCESS_IPC_REQUIRED,
-  ERR_CHILD_PROCESS_STDIO_MAXBUFFER,
-  ERR_INVALID_ARG_TYPE,
-  ERR_OUT_OF_RANGE,
-} = errorCodes;
 const { clearTimeout, setTimeout } = require('timers');
 const { getValidatedPath } = require('internal/fs/utils');
 const {
@@ -95,6 +95,8 @@ const {
 
 const MAX_BUFFER = 1024 * 1024;
 
+const permission = require('internal/process/permission');
+
 const isZOS = process.platform === 'os390';
 let addAbortListener;
 
@@ -139,7 +141,7 @@ function fork(modulePath, args = [], options) {
     validateObject(options, 'options');
   }
   options = { __proto__: null, ...options, shell: false };
-  options.execPath = options.execPath || process.execPath;
+  options.execPath ||= process.execPath;
   validateArgumentNullCheck(options.execPath, 'options.execPath');
 
   // Prepare arguments for fork:
@@ -237,8 +239,8 @@ function exec(command, options, callback) {
 }
 
 const customPromiseExecFunction = (orig) => {
-  return (...args) => {
-    const { promise, resolve, reject } = createDeferredPromise();
+  return assignFunctionName(orig.name, function(...args) {
+    const { promise, resolve, reject } = PromiseWithResolvers();
 
     promise.child = orig(...args, (err, stdout, stderr) => {
       if (err !== null) {
@@ -251,7 +253,7 @@ const customPromiseExecFunction = (orig) => {
     });
 
     return promise;
-  };
+  });
 };
 
 ObjectDefineProperty(exec, promisify.custom, {
@@ -273,9 +275,7 @@ function normalizeExecFileArgs(file, args, options, callback) {
     args = null;
   }
 
-  if (args == null) {
-    args = [];
-  }
+  args ??= [];
 
   if (typeof options === 'function') {
     callback = options;
@@ -283,9 +283,7 @@ function normalizeExecFileArgs(file, args, options, callback) {
     validateObject(options, 'options');
   }
 
-  if (options == null) {
-    options = kEmptyObject;
-  }
+  options ??= kEmptyObject;
 
   if (callback != null) {
     validateFunction(callback, 'callback');
@@ -393,8 +391,7 @@ function execFile(file, args, options, callback) {
     let stderr;
     if (encoding ||
       (
-        child.stdout &&
-        child.stdout.readableEncoding
+        child.stdout?.readableEncoding
       )) {
       stdout = ArrayPrototypeJoin(_stdout, '');
     } else {
@@ -402,8 +399,7 @@ function execFile(file, args, options, callback) {
     }
     if (encoding ||
       (
-        child.stderr &&
-        child.stderr.readableEncoding
+        child.stderr?.readableEncoding
       )) {
       stderr = ArrayPrototypeJoin(_stderr, '');
     } else {
@@ -418,13 +414,11 @@ function execFile(file, args, options, callback) {
     if (args?.length)
       cmd += ` ${ArrayPrototypeJoin(args, ' ')}`;
 
-    if (!ex) {
-      ex = genericNodeError(`Command failed: ${cmd}\n${stderr}`, {
-        code: code < 0 ? getSystemErrorName(code) : code,
-        killed: child.killed || killed,
-        signal: signal,
-      });
-    }
+    ex ||= genericNodeError(`Command failed: ${cmd}\n${stderr}`, {
+      code: code < 0 ? getSystemErrorName(code) : code,
+      killed: child.killed || killed,
+      signal: signal,
+    });
 
     ex.cmd = cmd;
     callback(ex, stdout, stderr);
@@ -544,6 +538,31 @@ function copyProcessEnvToEnv(env, name, optionEnv) {
   }
 }
 
+let permissionModelFlagsToCopy;
+
+function getPermissionModelFlagsToCopy() {
+  if (permissionModelFlagsToCopy === undefined) {
+    permissionModelFlagsToCopy = [...permission.availableFlags(), '--permission'];
+  }
+  return permissionModelFlagsToCopy;
+}
+
+function copyPermissionModelFlagsToEnv(env, key, args) {
+  // Do not override if permission was already passed to file
+  if (args.includes('--permission') || (env[key] && env[key].indexOf('--permission') !== -1)) {
+    return;
+  }
+
+  const flagsToCopy = getPermissionModelFlagsToCopy();
+  for (const arg of process.execArgv) {
+    for (const flag of flagsToCopy) {
+      if (arg.startsWith(flag)) {
+        env[key] = `${env[key] ? env[key] + ' ' + arg : arg}`;
+      }
+    }
+  }
+}
+
 function normalizeSpawnArguments(file, args, options) {
   validateString(file, 'file');
   validateArgumentNullCheck(file, 'file');
@@ -651,7 +670,8 @@ function normalizeSpawnArguments(file, args, options) {
     ArrayPrototypeUnshift(args, file);
   }
 
-  const env = options.env || process.env;
+  // Shallow copy to guarantee changes won't impact process.env
+  const env = options.env || { ...process.env };
   const envPairs = [];
 
   // process.env.NODE_V8_COVERAGE always propagates, making it possible to
@@ -671,6 +691,10 @@ function normalizeSpawnArguments(file, args, options) {
     copyProcessEnvToEnv(env, '_EDC_SUSV3', options.env);
   }
 
+  if (permission.isEnabled()) {
+    copyPermissionModelFlagsToEnv(env, 'NODE_OPTIONS', args);
+  }
+
   let envKeys = [];
   // Prototype values are intentionally included.
   for (const key in env) {
@@ -856,7 +880,7 @@ function spawnSync(file, args, options) {
 
   // We may want to pass data in on any given fd, ensure it is a valid buffer
   for (let i = 0; i < options.stdio.length; i++) {
-    const input = options.stdio[i] && options.stdio[i].input;
+    const input = options.stdio[i]?.input;
     if (input != null) {
       const pipe = options.stdio[i] = { ...options.stdio[i] };
       if (isArrayBufferView(input)) {
