@@ -1,18 +1,18 @@
 'use strict';
 
 const {
+  ArrayPrototypeMap,
   ObjectDefineProperty,
   ReflectApply,
-  ArrayPrototypeMap,
   Symbol,
 } = primordials;
 
 const {
+  DNSException,
   codes: {
     ERR_INVALID_ARG_TYPE,
     ERR_INVALID_ARG_VALUE,
   },
-  DNSException,
 } = require('internal/errors');
 
 const {
@@ -67,7 +67,7 @@ function resolver(bindingName) {
     req.callback = callback;
     req.hostname = name;
     req.oncomplete = onresolve;
-    req.ttl = !!(options && options.ttl);
+    req.ttl = !!(options?.ttl);
     const err = this._handle[bindingName](req, name);
     if (err) throw new DNSException(err, bindingName, name);
     if (hasObserver('dns')) {
