@@ -1,7 +1,6 @@
 'use strict';
 
 const {
-  ArrayBufferPrototypeGetByteLength,
   ArrayFrom,
   ArrayPrototypePush,
   ArrayPrototypeSlice,
@@ -9,12 +8,13 @@ const {
   Error,
   ErrorCaptureStackTrace,
   FunctionPrototypeCall,
+  NumberParseInt,
   ObjectDefineProperties,
   ObjectDefineProperty,
+  ObjectFreeze,
   ObjectGetOwnPropertyDescriptor,
   ObjectGetOwnPropertyDescriptors,
   ObjectGetPrototypeOf,
-  ObjectFreeze,
   ObjectPrototypeHasOwnProperty,
   ObjectSetPrototypeOf,
   ObjectValues,
@@ -27,18 +27,21 @@ const {
   RegExpPrototypeGetHasIndices,
   RegExpPrototypeGetIgnoreCase,
   RegExpPrototypeGetMultiline,
+  RegExpPrototypeGetSource,
   RegExpPrototypeGetSticky,
   RegExpPrototypeGetUnicode,
-  RegExpPrototypeGetSource,
   SafeMap,
   SafeSet,
   SafeWeakMap,
   SafeWeakRef,
+  StringPrototypeIncludes,
   StringPrototypeReplace,
+  StringPrototypeSlice,
   StringPrototypeToLowerCase,
   StringPrototypeToUpperCase,
   Symbol,
   SymbolFor,
+  SymbolPrototypeGetDescription,
   SymbolReplace,
   SymbolSplit,
 } = primordials;
@@ -46,27 +49,33 @@ const {
 const {
   codes: {
     ERR_NO_CRYPTO,
+    ERR_NO_TYPESCRIPT,
     ERR_UNKNOWN_SIGNAL,
   },
   isErrorStackTraceLimitWritable,
-  uvErrmapGet,
   overrideStackTrace,
+  uvErrmapGet,
 } = require('internal/errors');
 const { signals } = internalBinding('constants').os;
 const {
-  isArrayBufferDetached: _isArrayBufferDetached,
   guessHandleType: _guessHandleType,
+  defineLazyProperties,
   privateSymbols: {
     arrow_message_private_symbol,
     decorated_private_symbol,
   },
   sleep: _sleep,
 } = internalBinding('util');
-const { isNativeError } = internalBinding('types');
+const { isNativeError, isPromise } = internalBinding('types');
 const { getOptionValue } = require('internal/options');
+const assert = require('internal/assert');
 const { encodings } = internalBinding('string_decoder');
 
 const noCrypto = !process.versions.openssl;
+const noTypeScript = !process.versions.amaro;
+
+const isWindows = process.platform === 'win32';
+const isMacOS = process.platform === 'darwin';
 
 const experimentalWarnings = new SafeSet();
 
@@ -101,10 +110,12 @@ function getDeprecationWarningEmitter(
   shouldEmitWarning = () => true,
 ) {
   let warned = false;
-  return function() {
-    if (!warned && shouldEmitWarning()) {
+  return function(arg) {
+    if (!warned && shouldEmitWarning(arg)) {
       warned = true;
-      if (code !== undefined) {
+      if (code === 'ExperimentalWarning') {
+        process.emitWarning(msg, code, deprecated);
+      } else if (code !== undefined) {
         if (!codesWarned.has(code)) {
           const emitWarning = useEmitSync ?
             require('internal/process/warning').emitWarningSync :
@@ -135,6 +146,12 @@ function pendingDeprecate(fn, msg, code) {
     emitDeprecationWarning();
     return ReflectApply(fn, this, args);
   }
+
+  ObjectDefineProperty(deprecated, 'length', {
+    __proto__: null,
+    ...ObjectGetOwnPropertyDescriptor(fn, 'length'),
+  });
+
   return deprecated;
 }
 
@@ -173,6 +190,11 @@ function deprecate(fn, msg, code, useEmitSync) {
     deprecated.prototype = fn.prototype;
   }
 
+  ObjectDefineProperty(deprecated, 'length', {
+    __proto__: null,
+    ...ObjectGetOwnPropertyDescriptor(fn, 'length'),
+  });
+
   return deprecated;
 }
 
@@ -193,9 +215,17 @@ function assertCrypto() {
     throw new ERR_NO_CRYPTO();
 }
 
-// Return undefined if there is no match.
-// Move the "slow cases" to a separate function to make sure this function gets
-// inlined properly. That prioritizes the common case.
+function assertTypeScript() {
+  if (noTypeScript)
+    throw new ERR_NO_TYPESCRIPT();
+}
+
+/**
+ * Move the "slow cases" to a separate function to make sure this function gets
+ * inlined properly. That prioritizes the common case.
+ * @param {unknown} enc
+ * @returns {string | undefined} Returns undefined if there is no match.
+ */
 function normalizeEncoding(enc) {
   if (enc == null || enc === 'utf8' || enc === 'utf-8') return 'utf8';
   return slowCases(enc);
@@ -384,6 +414,10 @@ function getCWDURL() {
   return cachedURL;
 }
 
+function getSystemErrorMessage(err) {
+  return lazyUv().getErrorMessage(err);
+}
+
 function getSystemErrorName(err) {
   const entry = uvErrmapGet(err);
   return entry ? entry[0] : `Unknown system error ${err}`;
@@ -435,7 +469,10 @@ function promisify(original) {
           resolve(values[0]);
         }
       });
-      ReflectApply(original, this, args);
+      if (isPromise(ReflectApply(original, this, args))) {
+        process.emitWarning('Calling promisify on a function that returns a Promise is likely a mistake.',
+                            'DeprecationWarning', 'DEP0174');
+      }
     });
   }
 
@@ -538,17 +575,6 @@ function sleep(msec) {
   _sleep(msec);
 }
 
-function createDeferredPromise() {
-  let resolve;
-  let reject;
-  const promise = new Promise((res, rej) => {
-    resolve = res;
-    reject = rej;
-  });
-
-  return { promise, resolve, reject };
-}
-
 // https://heycam.github.io/webidl/#define-the-operations
 function defineOperation(target, name, method) {
   ObjectDefineProperty(target, name, {
@@ -592,46 +618,6 @@ function exposeGetterAndSetter(target, name, getter, setter = undefined) {
   });
 }
 
-function defineLazyProperties(target, id, keys, enumerable = true) {
-  const descriptors = { __proto__: null };
-  let mod;
-  for (let i = 0; i < keys.length; i++) {
-    const key = keys[i];
-    let lazyLoadedValue;
-    function set(value) {
-      ObjectDefineProperty(target, key, {
-        __proto__: null,
-        writable: true,
-        value,
-      });
-    }
-    ObjectDefineProperty(set, 'name', {
-      __proto__: null,
-      value: `set ${key}`,
-    });
-    function get() {
-      mod ??= require(id);
-      if (lazyLoadedValue === undefined) {
-        lazyLoadedValue = mod[key];
-        set(lazyLoadedValue);
-      }
-      return lazyLoadedValue;
-    }
-    ObjectDefineProperty(get, 'name', {
-      __proto__: null,
-      value: `get ${key}`,
-    });
-    descriptors[key] = {
-      __proto__: null,
-      configurable: true,
-      enumerable,
-      get,
-      set,
-    };
-  }
-  ObjectDefineProperties(target, descriptors);
-}
-
 function defineReplaceableLazyAttribute(target, id, keys, writable = true, check) {
   let mod;
   for (let i = 0; i < keys.length; i++) {
@@ -749,9 +735,9 @@ function SideEffectFreeRegExpPrototypeExec(regex, string) {
   return FunctionPrototypeCall(RegExpFromAnotherRealm.prototype.exec, regex, string);
 }
 
-const crossRelmRegexes = new SafeWeakMap();
-function getCrossRelmRegex(regex) {
-  const cached = crossRelmRegexes.get(regex);
+const crossRealmRegexes = new SafeWeakMap();
+function getCrossRealmRegex(regex) {
+  const cached = crossRealmRegexes.get(regex);
   if (cached) return cached;
 
   let flagString = '';
@@ -764,26 +750,17 @@ function getCrossRelmRegex(regex) {
   if (RegExpPrototypeGetSticky(regex)) flagString += 'y';
 
   const { RegExp: RegExpFromAnotherRealm } = getInternalGlobal();
-  const crossRelmRegex = new RegExpFromAnotherRealm(RegExpPrototypeGetSource(regex), flagString);
-  crossRelmRegexes.set(regex, crossRelmRegex);
-  return crossRelmRegex;
+  const crossRealmRegex = new RegExpFromAnotherRealm(RegExpPrototypeGetSource(regex), flagString);
+  crossRealmRegexes.set(regex, crossRealmRegex);
+  return crossRealmRegex;
 }
 
 function SideEffectFreeRegExpPrototypeSymbolReplace(regex, string, replacement) {
-  return getCrossRelmRegex(regex)[SymbolReplace](string, replacement);
+  return getCrossRealmRegex(regex)[SymbolReplace](string, replacement);
 }
 
 function SideEffectFreeRegExpPrototypeSymbolSplit(regex, string, limit = undefined) {
-  return getCrossRelmRegex(regex)[SymbolSplit](string, limit);
-}
-
-
-function isArrayBufferDetached(value) {
-  if (ArrayBufferPrototypeGetByteLength(value) === 0) {
-    return _isArrayBufferDetached(value);
-  }
-
-  return false;
+  return getCrossRealmRegex(regex)[SymbolSplit](string, limit);
 }
 
 /**
@@ -825,6 +802,59 @@ function setupCoverageHooks(dir) {
   return coverageDirectory;
 }
 
+// Returns the number of ones in the binary representation of the decimal
+// number.
+function countBinaryOnes(n) {
+  // Count the number of bits set in parallel, which is faster than looping
+  n = n - ((n >>> 1) & 0x55555555);
+  n = (n & 0x33333333) + ((n >>> 2) & 0x33333333);
+  return ((n + (n >>> 4) & 0xF0F0F0F) * 0x1010101) >>> 24;
+}
+
+function getCIDR(address, netmask, family) {
+  let ones = 0;
+  let split = '.';
+  let range = 10;
+  let groupLength = 8;
+  let hasZeros = false;
+  let lastPos = 0;
+
+  if (family === 'IPv6') {
+    split = ':';
+    range = 16;
+    groupLength = 16;
+  }
+
+  for (let i = 0; i < netmask.length; i++) {
+    if (netmask[i] !== split) {
+      if (i + 1 < netmask.length) {
+        continue;
+      }
+      i++;
+    }
+    const part = StringPrototypeSlice(netmask, lastPos, i);
+    lastPos = i + 1;
+    if (part !== '') {
+      if (hasZeros) {
+        if (part !== '0') {
+          return null;
+        }
+      } else {
+        const binary = NumberParseInt(part, range);
+        const binaryOnes = countBinaryOnes(binary);
+        ones += binaryOnes;
+        if (binaryOnes !== groupLength) {
+          if (StringPrototypeIncludes(binary.toString(2), '01')) {
+            return null;
+          }
+          hasZeros = true;
+        }
+      }
+    }
+  }
+
+  return `${address}/${ones}`;
+}
 
 const handleTypes = ['TCP', 'TTY', 'UDP', 'FILE', 'PIPE', 'UNKNOWN'];
 function guessHandleType(fd) {
@@ -834,6 +864,7 @@ function guessHandleType(fd) {
 
 class WeakReference {
   #weak = null;
+  // eslint-disable-next-line no-unused-private-class-members
   #strong = null;
   #refCount = 0;
   constructor(object) {
@@ -868,13 +899,40 @@ const encodingsMap = { __proto__: null };
 for (let i = 0; i < encodings.length; ++i)
   encodingsMap[encodings[i]] = i;
 
+/**
+ * Reassigns the .name property of a function.
+ * Should be used when function can't be initially defined with desired name
+ * or when desired name should include `#`, `[`, `]`, etc.
+ * @param {string | symbol} name
+ * @param {Function} fn
+ * @param {object} [descriptor]
+ * @returns {Function} the same function, renamed
+ */
+function assignFunctionName(name, fn, descriptor = kEmptyObject) {
+  if (typeof name !== 'string') {
+    const symbolDescription = SymbolPrototypeGetDescription(name);
+    assert(symbolDescription !== undefined, 'Attempted to name function after descriptionless Symbol');
+    name = `[${symbolDescription}]`;
+  }
+  return ObjectDefineProperty(fn, 'name', {
+    __proto__: null,
+    writable: false,
+    enumerable: false,
+    configurable: true,
+    ...ObjectGetOwnPropertyDescriptor(fn, 'name'),
+    ...descriptor,
+    value: name,
+  });
+}
+
 module.exports = {
   getLazy,
   assertCrypto,
+  assertTypeScript,
+  assignFunctionName,
   cachedResult,
   convertToValidSignal,
   createClassWrapper,
-  createDeferredPromise,
   decorateErrorStack,
   defineOperation,
   defineLazyProperties,
@@ -889,15 +947,18 @@ module.exports = {
   filterDuplicateStrings,
   filterOwnProperties,
   getConstructorOf,
+  getCIDR,
   getCWDURL,
   getInternalGlobal,
   getStructuredStack,
   getSystemErrorMap,
   getSystemErrorName,
+  getSystemErrorMessage,
   guessHandleType,
-  isArrayBufferDetached,
   isError,
   isUnderNodeModules,
+  isMacOS,
+  isWindows,
   join,
   lazyDOMException,
   lazyDOMExceptionClass,
@@ -937,4 +998,6 @@ module.exports = {
   setOwnProperty,
   pendingDeprecate,
   WeakReference,
+  isPendingDeprecation,
+  getDeprecationWarningEmitter,
 };
