@@ -9,10 +9,10 @@ if ((!common.hasCrypto) || (!common.hasIntl)) {
 
 common.skipIfEslintMissing();
 
-const RuleTester = require('../../tools/node_modules/eslint').RuleTester;
+const RuleTester = require('../../tools/eslint/node_modules/eslint').RuleTester;
 const rule = require('../../tools/eslint-rules/prefer-util-format-errors');
 
-new RuleTester({ parserOptions: { ecmaVersion: 6 } })
+new RuleTester()
   .run('prefer-util-format-errors', rule, {
     valid: [
       'E(\'ABC\', \'abc\');',
